package asd.td6;

import asd.base.ObjetAClef;

/**
 * Une classe pour l'exercice 4 de la feuille 6.
 */
public class ABRO {

    // ATTRIBUTS

    ObjetAClef racine; // la valeur du noeud
    ABRO gauche;       // le sous-arbre gauche
    ABRO droite;       // le sous-arbre droit

    // CONSTRUCTEURS

    /**
     * Construit un arbre binaire de recherche vide.
     */
    public ABRO() {
        gauche = null; // ou droite = null
    }

    /**
     * Construit un arbre binaire de recherche rduit 
     * une feuille de racine x.
     */
    public ABRO(ObjetAClef x) {
        racine = x;
        gauche = new ABRO();
        droite = new ABRO();
    }

    // METHODES PRIMITIVES

    /**
     * Teste si l'arbre est vide.
     */
    public boolean vide() {
        return gauche == null;
    }

    /**
     * Transforme l'arbre en feuille de valeur x, soit
     * un arbre binaire rduit  une feuille de racine x.
     * Prcondition : this est vide.
     */
    private void feuille(ObjetAClef x) {
        gauche = new ABRO();
        droite = new ABRO();
        racine = x;
    }

    /**
     * Remplace la racine et les sous-arbres gauche et droit de l'arbre
     * par la racine et les sous-arbres gauche et droit de l'arbre A.
     */
    private void devient(ABRO A) {
        racine = A.racine;
        gauche = A.gauche;
        droite = A.droite;
    }

    /**
     * Retire tous les lments de l'arbre.
     */
    private void couper() {
        gauche = null;
        droite = null; // inutile sauf peut-tre
        racine = null; // pour le GC Java !
    }

    // mthodes non primitives

    // QUESTION 1.1 : ajout d'un lment

    /**
     * Ajoute l'objet x  l'arbre. 
     * Complexit : O(h) avec lg(N) <= h <= N.
     */
    public void ajouter(ObjetAClef x) {
        if ( vide() )
            feuille(x);
        else
            if ( x.clef() <= racine.clef() )
                gauche.ajouter(x);
            else
                droite.ajouter(x);
    }

    // QUESTION 1.2 : recherche d'une occurence

    /**
     * Retourne la i-ime occurence de l'objet de clef k, ou null
     * si une telle occurence n'est pas prsente dans l'arbre.
     * Prcondition : i > 0.
     * Complexit : O(h) avec lg(N) <= h <= N.
     */
    public ObjetAClef rechercher(int k, int i) {
        if ( vide() )
            return null;
        if ( k < racine.clef() )
            return gauche.rechercher(k,i);
        if ( k > racine.clef() )
            return droite.rechercher(k,i);
        if ( i == 1 )
            return racine;
        return gauche.rechercher(k,i-1);
    }

    // QUESTION 1.3 : suppression d'une occurence

    /**
     * Supprime la i-ime occurence de l'objet de clef k de l'arbre.
     * Si une telle occurence n'est pas dans l'arbre, ne fait rien.
     * Prcondition : i > 0.
     * Complexit : O(h) avec lg(N) <= h <= N.
     */
    public void supprimer(int k, int i) {
        if ( ! vide() ) {
            if ( k < racine.clef() )
                gauche.supprimer(k,i);
            if ( k > racine.clef() )
                droite.supprimer(k,i);
            if ( k == racine.clef() )
                if ( i == 1 )
                    supprimerRacine();
                else
                    gauche.supprimer(k,i-1);
        }
    }

    /**
     * Supprime la racine de l'arbre, en la remplacant ventuellement
     * par le maximum du sous-arbre gauche.
     * Complexit : O(h) avec lg(N) <= h <= N.
     */
    private void supprimerRacine() {
        if ( gauche.vide() && droite.vide() )
            couper();
        else
            if ( gauche.vide() )
                devient(droite);
            else
                if ( droite.vide() )
                    devient(gauche);
                else
                    racine = gauche.supprimerMax();
    }

    /**
     * Supprime l'lment maximum de l'arbre et retourne cet lment.
     * Complexit : O(h) o h est la hauteur de A.
     */
    private ObjetAClef supprimerMax() {
        if ( droite.vide() ) {
            ObjetAClef max = racine;
            if ( gauche.vide() )
                couper();
            else {
                racine = gauche.racine;
                droite = gauche.droite;
                gauche = gauche.gauche;
            }
            return max;
        }
        return droite.supprimerMax();
    }

    // Affichage d'un l'ABRO

    /**
     * Affiche l'arbre sous une forme lisible.
     */
    public void afficher() {
        afficher("","");
    }

    /**
     * Affiche l'arbre sous une forme lisible avec r comme prfixe
     * de la chane correspondant  la racine et p comme prfixe
     * des autres chanes.
     */
    private void afficher(String r, String p) {
        if ( vide() ) {
            System.out.println(r);
        }
        else {
            String rs = racine.toString();
            System.out.println(r + rs);
            if ( ! gauche.vide() || ! droite.vide() ) {
                String rr = p + '|' + chaine('_',rs.length()) + ' ';
                droite.afficher(rr, p + '|' + chaine(' ',rs.length() + 1));
                System.out.println(p + '|');
                gauche.afficher(rr, p + chaine(' ',rs.length() + 2));
            }
        }
    }

    /**
     * Retourne une chane de longueur k ne contenant que le
     * caractre c.
     */
    private String chaine(char c, int k) {
        String s = "";
        for ( int i = 0; i < k; i++ ) {
            s += c;
        }
        return s;
    }
}
